/** Rill v0.0.497 https://www.npmjs.com/package/rill License: MIT*/
var Application, Request, Response, co, compose, expect, http, pathToReg, respond;

co = require("co");

pathToReg = require("path-to-regexp");

compose = require("./util/compose");

expect = require("./util/expect");

Request = require("./request");

Response = require("./response");

respond = require("./util/response");

http = (http = require("http"));


/*
 * Creates an isomorphic Rill application that can listen for browser and server page requests.
#
 * @constructor
 */

Application = (function() {
  var j, len, method, ref;

  function Application() {
    if (!(this instanceof Application)) {
      return new Application();
    }
    this.middleware = [];
  }


  /*
  	 * Starts the server or browser hijacker.
   */

  Application.prototype.listen = function() {
    var server;
    server = http.createServer(this.callback());
    server.listen.apply(server, arguments);
    return this;
  };


  /*
  	 * Creates a call back function for the current app that can handle incoming requests.
   */

  Application.prototype.callback = function() {
    var fn;
    fn = co.wrap(compose([respond].concat(this.middleware)));
    return function(req, res) {
      if (res == null) {
        res = {};
      }
      res.statusCode = 404;
      req = new Request(req);
      res = new Response(res);
      return fn.call({}, req, res).then(function() {
        return {
          redirected: Boolean(req.redirectURL)
        };
      })["catch"](function(err) {
        return console.error(err.stack || err);
      });
    };
  };


  /*
  	 * Attaches a midleware function to the current app that will be ran with requests.
  	 * @param {Generator} handler
   */

  Application.prototype.use = function(handler) {
    var ref;
    expect((handler != null ? (ref = handler.constructor) != null ? ref.name : void 0 : void 0) === "GeneratorFunction", "Can only use generator functions.");
    this.middleware.push(handler);
    return this;
  };


  /*
  	 * Simple syntactic sugar for functions that wish to modify the current application instance.
  	#
  	 * @param {Function} transformer
   */

  Application.prototype.setup = function(transformer) {
    expect(transformer(this) === this, "Setup failed, transformer function must return the modified application.");
    return this;
  };


  /*
  	 * Attaches a seperate rill application at a given path.
  	 * req.path will be updated to remove the provided path.
  	#
  	 * @param {Application} app
   */

  Application.prototype.mount = function(path, app) {
    var downstream;
    if (typeof path !== "string") {
      app = path;
      path = "/";
    }
    expect(path[0] === "/", "Mount path must begin with '/'.");
    expect(app instanceof Application, "Can only mount other Rill apps.");
    downstream = compose(app.middleware);
    if (path === "/") {
      this.use(downstream);
      return this;
    }
    path = pathToReg(path, [], {
      end: false
    });
    this.use(function*(req, res, upstream) {
      var originalPath;
      if (path.test(req.path)) {
        originalPath = req.path;
        (yield downstream.call(this, req, res, (function*() {
          (yield upstream);
          return req.path = req.path.replace(path, "");
        })()));
        return req.path = originalPath;
      } else {
        return (yield upstream);
      }
    });
    return this;
  };


  /*
  	 * Create functions for each http verb that will accept a path url and a middleware function to use.
  	#
  	 * @param {String} path
  	 * @param {Generator} middleware
   */

  ref = ["options", "head", "get", "put", "post", "delete"];
  for (j = 0, len = ref.length; j < len; j++) {
    method = ref[j];
    Application.prototype[method] = (function(method) {
      return function(path, middleware) {
        var keys;
        keys = [];
        path = pathToReg(path, keys);
        method = method.toUpperCase();
        this.use(function*(req, res, next) {
          var i, k, len1, match, name, optional, ref1, val;
          if ((req.method === method) && (match = req.path.match(path))) {
            req.params = {};
            for (i = k = 0, len1 = keys.length; k < len1; i = ++k) {
              ref1 = keys[i], name = ref1.name, optional = ref1.optional;
              val = match[i + 1];
              if (!(optional || (val != null))) {
                return (yield next);
              }
              req.params[name] = val;
            }
            return (yield middleware.call(this, req, res, next));
          } else {
            return (yield next);
          }
        });
        return this;
      };
    })(method);
  }

  return Application;

})();

module.exports = Application;
